#include "StdAfx.h"
#include "OptimusMini.h"

COptimusMini::COptimusMini() : m_bDeviceAlive(FALSE), m_bNowInitializing(FALSE)
{
#ifdef FEATURE_GAMMA
	DeviceGamma( 0.65 );
#endif // FEATURE_GAMMA

#ifdef FEATURE_COLOR_TEMPERATURE
	DeviceColorTemperature( 8500 );
#endif // FEATURE_COLOR_TEMPERATURE
}

COptimusMini::~COptimusMini()
{
	m_objPort.Close();
}

BOOL COptimusMini::InitDevice()
{
	DEBUG(DEBUG_INIT,printf( "Init begin\n" ));

	BYTE cValue;
	unsigned long lRes;

	m_objPort.SetTimeout( 0 ); // by default read is blocking, unblocking it

	do
	{
		cValue=0; // can be altered by read operation
		int iCounter=0;

		do {
			// sending 0 until device response something
			if( !m_objPort.WriteFile( &cValue, 1, &lRes ) )
			{
				ReportProblem( PROBLEM_WRITE_ERROR );
				return FALSE;
			}

			if ( lRes==0 || iCounter>300 )
			{
				ReportProblem( PROBLEM_WRITE_TIMEOUT );
				return FALSE;
			}

			if ( iCounter == COMMAND_LENGTH - 6 )
			{
				m_objPort.SetTimeout(1); // start sleeping to assure getting device response before sending next zero.
			}

			if( !m_objPort.ReadFile( &cValue, 1, &lRes ) )
			{
				ReportProblem( PROBLEM_READ_ERROR );
				return FALSE;
			}

			iCounter++;
	} 
		while ( lRes==0 );

		DEBUG(DEBUG_INIT,printf( "Init counter %d\n", iCounter ));
	} 
	while( cValue!=0 );

	m_objPort.SetTimeout( PORT_TIMEOUT ); // setting normal timeout

	if( !m_objPort.ReadFile( &cValue, 1, &lRes ) )
	{
		ReportProblem( PROBLEM_READ_ERROR );
		return FALSE;
	}

	if ( lRes==0 )
	{
		ReportProblem( PROBLEM_READ_TIMEOUT );
		return FALSE;
	}

	DEBUG(DEBUG_INIT,printf( "Init success\n" ));

	OnDeviceStateChanged( TRUE );

	return TRUE;
}


int COptimusMini::SendBytes( BYTE *command, BYTE checksum )
{
	if( !m_bDeviceAlive )
	{
		return COMMAND_FAILED;
	}

	unsigned long lRes;

	command[COMMAND_LENGTH-1]=checksum;

	if( !m_objPort.WriteFile( command,COMMAND_LENGTH, &lRes ) || lRes!=COMMAND_LENGTH )
	{
		ReportProblem( PROBLEM_WRITE_ERROR );
		return COMMAND_FAILED;
	}

	BYTE cValue = 0;
	BYTE cDeviceId;
	int iState=0;

	for(;;)
	{
		if( !m_objPort.ReadFile( &cValue, 1, &lRes ) )
		{
			ReportProblem( PROBLEM_READ_ERROR );
			return COMMAND_FAILED;
		}

		if (lRes==0)
		{
			ReportProblem( PROBLEM_READ_TIMEOUT );
			return COMMAND_FAILED;
		}

		switch (iState)
		{
		case 0:

			if (cValue == 0 || cValue == 1 || command[0] == 8 && cValue == 2 )
			{
				iState = cValue + 1;
			} 
			else 
			{
				if (cValue == checksum)
				{
					return 0;
				}
			}
			break;

		case 1:
			if( checksum != cValue )
			{
				DEBUG(DEBUG_NOTIFY,printf( "SendBytes checksum mismatch: %d vs %d (received)\n", checksum, cValue ));
				ReportProblem( PROBLEM_CHECKSUM_MISMATCH );
			} 
			return 0;

		case 2:
			if( cValue > 0 && cValue < 4 )
			{
				OnKeyDown( cValue );
			}
			else 
			{
				DEBUG(DEBUG_NOTIFY,printf( "SendBytes wrong key: %d\n", cValue ));
				ReportProblem( PROBLEM_INVALID_KEYDOWN );
			}
			iState = 0;
			break;

		case 3:
			cDeviceId = cValue;
			iState++;
			break;

		case 4:
			return ((int)cDeviceId << 8 ) + cValue;
		}
	}

	return 0;
}

int COptimusMini::SendCommand( BYTE c1, BYTE c2, BYTE c3, BYTE c4 )
{
	if( !m_bDeviceAlive )
	{
		return COMMAND_FAILED;
	}

	DEBUG(DEBUG_INTERNAL,printf( "SendCommand: %d, %d\n", c1, c2 ));

	BYTE buf[COMMAND_LENGTH] = {c1,c2,c3,c4};
	BYTE sum=buf[0]+buf[1]+buf[2]+buf[3];

	for(int i=0;i<PORT_RETRY;i++)
	{
		int res = SendBytes( buf, sum );

		if (res==0 || c1==8)
		{
			DEBUG(DEBUG_INTERNAL_DETAIL,printf( "SendCommand done: %d, %d\n", c1, c2 ));
			return res;
		}
	}

	DEBUG(DEBUG_WARNING,printf( "SendCommand: %d, %d failed\n", c1, c2 ));

	return COMMAND_FAILED;
}

int COptimusMini::SendLine(int button, int line, BYTE *value)
{
	BYTE buf[COMMAND_LENGTH] = { 1, button, line*192>>8, line*192 };
	BYTE checksum = buf[0] + buf[1] + buf[2] + buf[3];
	BYTE *ptr = buf + 4;

	for (BYTE *rgb=value+SCREEN_SIZE*3-6;rgb>=value;rgb-=6)
	{
		BYTE	r1 = rgb[0], g1 = rgb[1], b1 = rgb[2],
			r2 = rgb[3], g2 = rgb[4], b2 = rgb[5];

#ifdef FEATURE_GAMMA

		r1 = m_cGammaTable[r1];
		g1 = m_cGammaTable[g1];
		b1 = m_cGammaTable[b1];

		r2 = m_cGammaTable[r2];
		g2 = m_cGammaTable[g2];
		b2 = m_cGammaTable[b2];

#endif // FEATURE_GAMMA

#ifdef FEATURE_COLOR_TEMPERATURE

		r1 = m_crColorTable[r1];
		g1 = m_cgColorTable[g1];
		b1 = m_cbColorTable[b1];

		r2 = m_crColorTable[r2];
		g2 = m_cgColorTable[g2];
		b2 = m_cbColorTable[b2];

#endif // FEATURE_COLOR_TEMPERATURE

#if 1 // fix colors like 040404 "greening" due to extra green bit

		if (g1<8){
			g1=0;
		}

		if (g2<8){
			g2=0;
		}

#endif
		ptr[0]=(b2&0xF8) + (g2>>5);
		ptr[1]=(r2>>3) + ((g2&0xFC)<<3);

		ptr[2]=(b1&0xF8) + (g1>>5);
		ptr[3]=(r1>>3) + ((g1&0xFC)<<3);

		checksum+=ptr[0]+ptr[1]+ptr[2]+ptr[3];
		ptr+=4;
	}

	for(int i=0;i<PORT_RETRY;i++)
	{
		if( SendBytes( buf,checksum ) == 0 )
		{
			return 0;
		}
	}

	DEBUG(DEBUG_WARNING,printf( "SendLine: %d, %d failed\n", button, line ));
	return COMMAND_FAILED;
}
	

int COptimusMini::ShowPicture( int p_nButton, LPBYTE p_lpData, int p_nFirstLine, int p_nLastLine, BOOL p_bShow ){
	
	if( p_lpData == NULL || p_nButton < 1 || p_nButton > 3 )
	{
		return COMMAND_FAILED;
	}

	DEBUG(DEBUG_INTERNAL,printf( "ShowPicture begin, %d\n", p_nButton ));

	// FirstLine and LastLine counted for buffer lines;
	// Device uses reverse order - first line is bottom line, inverting.

	for( int i=SCREEN_SIZE-p_nLastLine; i<SCREEN_SIZE-p_nFirstLine; i++ )
	{
		if (SendLine( p_nButton, i, p_lpData + (SCREEN_SIZE-1-i)*SCREEN_SIZE*3) != 0)
		{
			return COMMAND_FAILED;
		}
	}

	if (p_bShow)
	{
		SendCommand( 4, p_nButton );
	}

	DEBUG(DEBUG_INTERNAL,printf( "ShowPicture end\n" ));
	return 0;
}

BOOL COptimusMini::Open()
{
	DEBUG(DEBUG_INIT,printf( "Detecting device begin\n" ));

	TCHAR *szPorts;
	TCHAR *szDevParameters;

	//WIN98: HKEY_LOCAL_MACHINE\Enum\USBPORTS\VID_067B&PID_2303
	//WINNT: HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Enum\USB\Vid_067b&Pid_2303

	if( ::GetVersion() < 0x80000000 ) // NT
	{
		szPorts="SYSTEM\\CurrentControlSet\\Enum\\USB\\Vid_067b&Pid_2303";
		szDevParameters="\\Device Parameters";
	}
	else // 9x
	{
		szPorts="Enum\\USBPORTS\\VID_067B&PID_2303";
		szDevParameters="";
	}

	HKEY hPortsKey;
	if( ::RegOpenKeyEx( HKEY_LOCAL_MACHINE, szPorts, 0, KEY_ENUMERATE_SUB_KEYS, &hPortsKey ) != ERROR_SUCCESS )
	{
		ReportProblem( PROBLEM_DRIVER_NOT_FOUND );
		return FALSE;
	}

	for( int n=0;; n++ )
	{
		TCHAR szKeyName[100];
		DWORD dwKeyNameSize = sizeof(szKeyName);

		if( ::RegEnumKeyEx( hPortsKey, n, szKeyName, &dwKeyNameSize, NULL, NULL, NULL, NULL ) != ERROR_SUCCESS )
		{
			break;
		}

		TCHAR szPortValueName[200] = "";
		strcat( szPortValueName, szKeyName );
		strcat( szPortValueName, szDevParameters );

		HKEY hDevParams;

		if ( ::RegOpenKeyEx( hPortsKey, szPortValueName, 0, KEY_QUERY_VALUE, &hDevParams ) != ERROR_SUCCESS )
		{
			continue;
		}

		BYTE szValueData[100];
		DWORD dwValueDataSize = sizeof(szValueData);

		LONG gle = ::RegQueryValueEx( hDevParams, "PortName", NULL, NULL, szValueData, &dwValueDataSize );

		::RegCloseKey( hDevParams );

		if( gle == ERROR_SUCCESS && m_objPort.Open( (LPSTR)szValueData ) )
		{
			break;
		}
	}

	::RegCloseKey( hPortsKey );

	return m_objPort.IsOpened();
}

BOOL COptimusMini::Init()
{
	return Open() && InitDevice();
}

void COptimusMini::Close()
{
	m_objPort.Close();
	m_bDeviceAlive = FALSE; // avoid calling callback function
}

void COptimusMini::OnKeyDown( BYTE p_cValue )
{
	DEBUG(DEBUG_NOTIFY,printf( "OnKeyDown: %d \n", p_cValue ));
}

void COptimusMini::OnDeviceStateChanged( BOOL p_bState )
{
	if (m_bDeviceAlive != p_bState )
	{
		m_bDeviceAlive = p_bState;
		DEBUG(DEBUG_NOTIFY,printf( "OnDeviceStateChanged: %s \n", p_bState ? "alive" : "dead" ));
	}
}

void COptimusMini::ReportProblem( int p_nErrorCode )
{
	DEBUG(DEBUG_WARNING,printf( "Problem: %d, gle=%d\n", p_nErrorCode, ::GetLastError() ));

	if( (p_nErrorCode & 0x10) > 0 )
	{
		if( m_bNowInitializing )
		{
			m_objPort.Close();
			DEBUG(DEBUG_WARNING,printf( "re-Init() failed\n" ));
			OnDeviceStateChanged( FALSE );
		}
		else
		{
			/* driver hangs if we attempt to reinit device after read timeout, thus device should be reconnected */
			switch (p_nErrorCode)
			{
			case PROBLEM_CHECKSUM_MISMATCH:
				InitDevice();
				return;

			case PROBLEM_READ_TIMEOUT:
				DEBUG(DEBUG_WARNING,printf("Read timeout, possible Optimus reconnect\n"));

			default:
				m_objPort.Close();
				Sleep(500);
				m_objPort.Open( NULL );
				Sleep(500);
			}

			m_bNowInitializing = TRUE;
			InitDevice();
			m_bNowInitializing = FALSE;
		}
	}
}

#ifdef FEATURE_GAMMA

#include <math.h>

void COptimusMini::DeviceGamma( double p_dGamma )
{
	if (p_dGamma < 0.1)
	{
		p_dGamma = 0.1;
	}

	if (p_dGamma > 10)
	{
		p_dGamma = 10;
	}

	for (int i=0; i<256; i++)
	{
		m_cGammaTable[i]= (byte)floor( pow( i/255.0, 1/p_dGamma ) * 255 + 0.5 );
	}
}

#endif // FEATURE_GAMMA

#ifdef FEATURE_COLOR_TEMPERATURE

#include "BBRcolor.h"

void COptimusMini::DeviceColorTemperature( int p_iTemperature )
{
	int idx = (p_iTemperature-1000+50)/100;

	if (idx<0 || idx >= sizeof(mWritePoint)/sizeof(COLORTEMPERATURE) )
	{
		return;
	}

	int r=mWritePoint[idx].r,
		g=mWritePoint[idx].g,
		b=mWritePoint[idx].b;

	for (int i=0; i<256; i++)
	{
		 m_crColorTable[i]= i*r/255;
		 m_cgColorTable[i]= i*g/255;
		 m_cbColorTable[i]= i*b/255;
	}
}

#endif // FEATURE_COLOR_TEMPERATURE
