#pragma once

#include "Port.h"
#include <queue>

// OLED display has 96 pixes on each side
#define SCREEN_SIZE 96
// Space required for RGB values for each OLED pixel
#define SCREEN_BUFFER_SIZE (SCREEN_SIZE*SCREEN_SIZE*3)
// Command is command id, display number, byte offset, row data (2 bytes for each pixel), checksum
#define COMMAND_LENGTH (1+1+2+SCREEN_SIZE*2+1)

// **************** Features to be enabled or disabled in compile time ****************

// Enables gamma correction support with default gamma set to 0,65
#define FEATURE_GAMMA
// Enables color temperature correction support with default temperature set to 8500
#define FEATURE_COLOR_TEMPERATURE

// **************** Error codes are sent to ReportProblem function ****************

#define COMMAND_FAILED -1

/* 0x10 bit means device should be reinited */

#define PROBLEM_WRITE_ERROR			0x10
#define PROBLEM_WRITE_TIMEOUT		0x11
#define PROBLEM_READ_ERROR			0x12
#define PROBLEM_READ_TIMEOUT		0x13
#define PROBLEM_CHECKSUM_MISMATCH	0x14
#define PROBLEM_INVALID_KEYDOWN		0x01
#define PROBLEM_DRIVER_NOT_FOUND	0x02

// **************** Optimus mini low level communication class ****************

// Low-level device communication. Not thread safe.
// Can execute commands, show images, accept keypresses

class COptimusMini
{
protected:

	CPort m_objPort;					// read/write functions
	BOOL m_bNowInitializing;

public:
	
	BOOL m_bDeviceAlive;

	COptimusMini();
	~COptimusMini();

#ifdef FEATURE_GAMMA
	BYTE m_cGammaTable[256];
	void DeviceGamma( double p_dGamma );
#endif

#ifdef FEATURE_COLOR_TEMPERATURE
	BYTE m_crColorTable[256], m_cgColorTable[256], m_cbColorTable[256];
	void DeviceColorTemperature( int p_iTemperature );
#endif

	// scans all ports with VID_067B&PID_2303 device driver
	BOOL Open();

	// inits the device and calls OnDeviceStateChanged(TRUE) on success
	BOOL Init();

	// closes open port and m_bDeviceAlive=FALSE
	void Close();

	// prepares command buffer and executes SendBytes
	int SendCommand( BYTE c1, BYTE c2, BYTE c3 = 0, BYTE c4 = 0 );

	// shows picture on specified button
	int ShowPicture( int p_nButton, LPBYTE p_lpData, int p_nFirstLine = 0, int p_nLastLine = SCREEN_SIZE, BOOL p_bShow = TRUE );

	// returns active port
	char *GetPortName(){ return m_objPort.GetPortName(); };

protected:

	// attempts to init device on specified port
	BOOL InitDevice();

	// low level command execution function
	int SendBytes( BYTE *command, BYTE checksum );
	
	// image display support function
	int SendLine(int button, int line, BYTE *value);

	// errors handling function
	virtual void ReportProblem( int p_nErrorCode );

	virtual void OnKeyDown( BYTE p_cValue );
	virtual void OnDeviceStateChanged( BOOL p_bState );
};
